#!/usr/local/bin/python
# encoding: utf-8
"""
*Iterate through large line-based files in batches of lines*

:Author:
    David Young

:Date Created:
    December  4, 2017
"""
################# GLOBAL IMPORTS ####################
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
import codecs


class fileChunker(object):
    """
    *The fileChunker iterator - iterate over large line-based files to reduce memory footprint*

    **Key Arguments:**
        - ``filepath`` -- path to the large file to iterate over
        - ``batchSize`` -- size of the chunks to return in lines

    **Usage:**

        To setup your logger, settings and database connections, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_). 

        To initiate a fileChunker iterator and then process the file in batches of 100000 lines, use the following:

        .. code-block:: python 

            from fundamentals.files import fileChunker
            fc = fileChunker(
                filepath="/path/to/large/file.csv",
                batchSize=100000
            )
            for i in fc:
                print len(i)
    """

    def __init__(self, filepath, batchSize):
        self.filepath = filepath
        self.batchSize = batchSize

        try:
            self.readFile = codecs.open(
                self.filepath, encoding='utf-8', mode='r')
        except IOError, e:
            message = 'could not open the file %s' % (self.filepath,)
            raise IOError(message)

    def __iter__(self): return self

    def next(self):
        batch = []
        for lines in range(self.batchSize):
            l = self.readFile.readline()
            if len(l):
                batch.append(l)
        if len(batch) == 0:
            raise StopIteration

        return batch



#!/usr/local/bin/python
# encoding: utf-8
"""
*Convert a python list of dictionaries to pretty csv output*

:Author:
    David Young

:Date Created:
    June 28, 2016
"""
################# GLOBAL IMPORTS ####################
import sys
import os
import io
import csv
from decimal import Decimal
from datetime import datetime
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from fundamentals.mysql import convert_dictionary_to_mysql_table


def list_of_dictionaries_to_mysql_inserts(
        log,
        datalist,
        tableName):
    """Convert a python list of dictionaries to pretty csv output

    **Key Arguments:**
        - ``log`` -- logger
        - ``datalist`` -- a list of dictionaries
        - ``tableName`` -- the name of the table to create the insert statements for

    **Return:**
        - ``output`` -- the mysql insert statements (as a string)

    **Usage:**

        .. code-block:: python 

            from fundamentals.files import list_of_dictionaries_to_mysql_inserts
            mysqlInserts = list_of_dictionaries_to_mysql_inserts(
                log=log,
                datalist=dataList,
                tableName="my_new_table"
            )
            print mysqlInserts

        this output the following:

        .. code-block:: plain

            INSERT INTO `testing_table` (a_newKey,and_another,dateCreated,uniqueKey2,uniquekey1) VALUES ("cool" ,"super cool" ,"2016-09-14T13:17:26" ,"burgers" ,"cheese")  ON DUPLICATE KEY UPDATE  a_newKey="cool", and_another="super cool", dateCreated="2016-09-14T13:17:26", uniqueKey2="burgers", uniquekey1="cheese" ;
            ...
            ...
    """
    log.debug('starting the ``list_of_dictionaries_to_mysql_inserts`` function')

    if not len(datalist):
        return "NO MATCH"

    inserts = []

    for d in datalist:
        insertCommand = convert_dictionary_to_mysql_table(
            log=log,
            dictionary=d,
            dbTableName="testing_table",
            uniqueKeyList=[],
            dateModified=False,
            returnInsertOnly=True,
            replace=True,
            batchInserts=False
        )
        inserts.append(insertCommand)

    output = ";\n".join(inserts) + ";"

    log.debug('completed the ``list_of_dictionaries_to_mysql_inserts`` function')
    return output



#!/usr/local/bin/python
# encoding: utf-8
"""
*List the contents of a directory recursively*

:Author:
    David Young

:Date Created:
    June 17, 2016
"""
################# GLOBAL IMPORTS ####################
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools


def recursive_directory_listing(
    log,
    baseFolderPath,
    whatToList="all"
):
    """*list directory contents recursively.*

    Options to list only files or only directories.

    **Key Arguments:**
        - ``log`` -- logger
        - ``baseFolderPath`` -- path to the base folder to list contained files and folders recursively
        - ``whatToList`` -- list files only, durectories only or all [ "files" | "dirs" | "all" ]

    **Return:**
        - ``matchedPathList`` -- the matched paths

    **Usage:**

        .. code-block:: python 

            from fundamentals.files import recursive_directory_listing
            theseFiles = recursive_directory_listing(
                log,
                baseFolderPath="/tmp"
            )

            # OR JUST FILE  

            from fundamentals.files import recursive_directory_listing
            theseFiles = recursive_directory_listing(
                log,
                baseFolderPath="/tmp",
                whatToList="files"
            )


            # OR JUST FOLDERS   

            from fundamentals.files import recursive_directory_listing
            theseFiles = recursive_directory_listing(
                log,
                baseFolderPath="/tmp",
                whatToList="dirs"
            )
            print theseFiles        
    """
    log.debug('starting the ``recursive_directory_listing`` function')

    ## VARIABLES ##
    matchedPathList = []
    parentDirectoryList = [baseFolderPath, ]

    count = 0
    while os.listdir(baseFolderPath) and count < 20:
        count += 1

        while len(parentDirectoryList) != 0:
            childDirList = []
            for parentDir in parentDirectoryList:
                try:
                    thisDirList = os.listdir(parentDir)
                except Exception, e:
                    log.error(e)
                    continue

                for d in thisDirList:
                    fullPath = os.path.join(parentDir, d)

                    if whatToList is "all":
                        matched = True
                    elif whatToList is "dirs":
                        matched = os.path.isdir(fullPath)
                    elif whatToList is "files":
                        matched = os.path.isfile(fullPath)
                    else:
                        log.error(
                            'cound not list files in %s, `whatToList` variable incorrect: [ "files" | "dirs" | "all" ]' % (baseFolderPath,))
                        sys.exit(0)

                    if matched:
                        matchedPathList.append(fullPath)

                    # UPDATE DIRECTORY LISTING
                    if os.path.isdir(fullPath):
                        childDirList.append(fullPath)

                parentDirectoryList = childDirList

    log.debug('completed the ``recursive_directory_listing`` function')
    return matchedPathList



#!/usr/local/bin/python
# encoding: utf-8
"""
*Given a yaml settings file containing database connection details, setup and return the db connector*

:Author:
    David Young

:Date Created:
    June 21, 2016
"""
################# GLOBAL IMPORTS ####################
import sys
import os
import yaml
os.environ['TERM'] = 'vt100'
from fundamentals import tools


def setup_database_connection(
        pathToYamlFile):
    """*Start a database connection using settings in yaml file* 

    Given the location of a YAML dictionary containing database credientials, this function will setup and return the connection*

    **Key Arguments:**
        - ``pathToYamlFile`` -- path to the YAML dictionary.

    **Return:**
        - ``dbConn`` -- connection to the MySQL database.

    **Usage:**

        The settings file should be in this form, with all keyword values set:

        .. code-block:: yaml

            db: dryx_unit_testing
            host: localhost
            user: unittesting
            password: utpass

        And here's how to generate the connection object:

        .. code-block:: python 

            from fundamentals.mysql import setup_database_connection
            dbConn = setup_database_connection(
                pathToYamlFile=pathToMyYamlFile
            )
    """
    import sys
    import logging
    import pymysql as ms

    # IMPORT THE YAML CONNECTION DICTIONARY
    try:
        logging.info(
            'importing the yaml database connection dictionary from ' + pathToYamlFile)
        stream = file(pathToYamlFile, 'r')
        connDict = yaml.load(stream)
    except:
        logging.critical(
            'could not load the connect dictionary from ' + pathToYamlFile)
        sys.exit(1)
    # ESTABLISH A DB CONNECTION
    try:
        logging.info('connecting to the ' + connDict[
                     'db'] + ' database on ' + connDict['host'])
        dbConn = ms.connect(
            host=connDict['host'],
            user=connDict['user'],
            passwd=connDict['password'],
            db=connDict['db'],
            use_unicode=True,
            charset='utf8',
            local_infile=1,
            client_flag=ms.constants.CLIENT.MULTI_STATEMENTS,
            connect_timeout=36000
        )
        dbConn.autocommit(True)
    except Exception as e:
        logging.critical('could not connect to the ' + connDict['db'] + ' database on ' + connDict['host'] + ' : '
                         + str(e) + '\n')
    return dbConn


# use the tab-trigger below for new function
# xt-def-function



#!/usr/local/bin/python
# encoding: utf-8
"""
*Given a list of dictionaries this function will insert each dictionary as a row into the given database table*

:Author:
    David Young

:Date Created:
    June 21, 2016
"""
################# GLOBAL IMPORTS ####################
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from fundamentals.mysql import convert_dictionary_to_mysql_table, writequery
from fundamentals.fmultiprocess import fmultiprocess
import time
import re
from fundamentals.mysql.database import database
import pandas as pd
from datetime import datetime
import numpy as np


count = 0
totalCount = 0
globalDbConn = False
sharedList = []


def insert_list_of_dictionaries_into_database_tables(
        dbConn,
        log,
        dictList,
        dbTableName,
        uniqueKeyList=[],
        dateModified=False,
        dateCreated=True,
        batchSize=2500,
        replace=False,
        dbSettings=False):
    """insert list of dictionaries into database tables

    **Key Arguments:**
        - ``dbConn`` -- mysql database connection
        - ``log`` -- logger
        - ``dictList`` -- list of python dictionaries to add to the database table
        - ``dbTableName`` -- name of the database table
        - ``uniqueKeyList`` -- a list of column names to append as a unique constraint on the database
        - ``dateModified`` -- add the modification date as a column in the database
        - ``dateCreated`` -- add the created date as a column in the database
        - ``batchSize`` -- batch the insert commands into *batchSize* batches
        - ``replace`` -- repalce row if a duplicate is found
        - ``dbSettings`` -- pass in the database settings so multiprocessing can establish one connection per process (might not be faster)

    **Return:**
        - None

    **Usage:**

        .. code-block:: python

            from fundamentals.mysql import insert_list_of_dictionaries_into_database_tables
            insert_list_of_dictionaries_into_database_tables(
                dbConn=dbConn,
                log=log,
                dictList=dictList,
                dbTableName="test_insert_many",
                uniqueKeyList=["col1", "col3"],
                dateModified=False,
                batchSize=2500
            )
    """

    log.debug(
        'completed the ````insert_list_of_dictionaries_into_database_tables`` function')

    global count
    global totalCount
    global globalDbConn
    global sharedList

    reDate = re.compile('^[0-9]{4}-[0-9]{2}-[0-9]{2}T')

    if dbSettings:
        globalDbConn = dbSettings
    else:
        globalDbConn = dbConn

    if len(dictList) == 0:
        log.warning(
            'the dictionary to be added to the database is empty' % locals())
        return None

    if len(dictList):
        convert_dictionary_to_mysql_table(
            dbConn=dbConn,
            log=log,
            dictionary=dictList[0],
            dbTableName=dbTableName,
            uniqueKeyList=uniqueKeyList,
            dateModified=dateModified,
            reDatetime=reDate,
            replace=replace,
            dateCreated=dateCreated)
        dictList = dictList[1:]

    dbConn.autocommit(False)

    if len(dictList):

        total = len(dictList)
        batches = int(total / batchSize)

        start = 0
        end = 0
        sharedList = []
        for i in range(batches + 1):
            end = end + batchSize
            start = i * batchSize
            thisBatch = dictList[start:end]
            sharedList.append((thisBatch, end))

        totalCount = total + 1
        ltotalCount = totalCount

        print "Starting to insert %(ltotalCount)s rows into %(dbTableName)s" % locals()

        if dbSettings == False:

            fmultiprocess(
                log=log,
                function=_insert_single_batch_into_database,
                inputArray=range(len(sharedList)),
                dbTableName=dbTableName,
                uniqueKeyList=uniqueKeyList,
                dateModified=dateModified,
                replace=replace,
                batchSize=batchSize,
                reDatetime=reDate,
                dateCreated=dateCreated
            )

        else:
            fmultiprocess(log=log, function=_add_dictlist_to_database_via_load_in_file,
                          inputArray=range(len(sharedList)), dbTablename=dbTableName,
                          dbSettings=dbSettings, dateModified=dateModified)

        sys.stdout.write("\x1b[1A\x1b[2K")
        print "%(ltotalCount)s / %(ltotalCount)s rows inserted into %(dbTableName)s" % locals()

    log.debug(
        'completed the ``insert_list_of_dictionaries_into_database_tables`` function')
    return None


def _insert_single_batch_into_database(
        batchIndex,
        log,
        dbTableName,
        uniqueKeyList,
        dateModified,
        replace,
        batchSize,
        reDatetime,
        dateCreated):
    """*summary of function*

    **Key Arguments:**
        - ``batchIndex`` -- the index of the batch to insert
        - ``dbConn`` -- mysql database connection
        - ``log`` -- logger

    **Return:**
        - None

    **Usage:**
        .. todo::

            add usage info
            create a sublime snippet for usage

        .. code-block:: python 

            usage code            
    """
    log.debug('starting the ``_insert_single_batch_into_database`` function')

    global totalCount
    global globalDbConn
    global sharedList

    batch = sharedList[batchIndex]

    reDate = reDatetime

    if isinstance(globalDbConn, dict):
        # SETUP ALL DATABASE CONNECTIONS

        dbConn = database(
            log=log,
            dbSettings=globalDbConn,
            autocommit=False
        ).connect()
    else:
        dbConn = globalDbConn

    count = batch[1]
    if count > totalCount:
        count = totalCount
    ltotalCount = totalCount

    inserted = False
    while inserted == False:

        if not replace:
            insertVerb = "INSERT IGNORE"
        else:
            insertVerb = "INSERT IGNORE"

        uniKeys = set().union(*(d.keys() for d in batch[0]))
        tmp = []
        tmp[:] = [m.replace(" ", "_").replace(
            "-", "_") for m in uniKeys]
        uniKeys = tmp

        myKeys = '`,`'.join(uniKeys)
        vals = [tuple([None if d[k] in ["None", None] else str(d[k])
                       for k in uniKeys]) for d in batch[0]]
        valueString = ("%s, " * len(vals[0]))[:-2]
        insertCommand = insertVerb + """ INTO `""" + dbTableName + \
            """` (`""" + myKeys + """`, dateCreated) VALUES (""" + \
            valueString + """, NOW())"""

        if not dateCreated:
            insertCommand = insertCommand.replace(
                ", dateCreated)", ")").replace(", NOW())", ")")

        dup = ""
        if replace:
            dup = " ON DUPLICATE KEY UPDATE "
            for k in uniKeys:
                dup = """%(dup)s %(k)s=values(%(k)s),""" % locals()
            dup = """%(dup)s updated=1, dateLastModified=NOW()""" % locals()

        insertCommand = insertCommand + dup

        insertCommand = insertCommand.replace('\\""', '\\" "')
        insertCommand = insertCommand.replace('""', "null")
        insertCommand = insertCommand.replace('"None"', 'null')

        message = ""
        # log.debug('adding new data to the %s table; query: %s' %
        # (dbTableName, addValue))
        try:
            message = writequery(
                log=log,
                sqlQuery=insertCommand,
                dbConn=dbConn,
                Force=True,
                manyValueList=vals
            )
        except:
            theseInserts = []
            for aDict in batch[0]:

                insertCommand, valueTuple = convert_dictionary_to_mysql_table(
                    dbConn=dbConn,
                    log=log,
                    dictionary=aDict,
                    dbTableName=dbTableName,
                    uniqueKeyList=uniqueKeyList,
                    dateModified=dateModified,
                    returnInsertOnly=True,
                    replace=replace,
                    reDatetime=reDate,
                    skipChecks=True
                )
                theseInserts.append(valueTuple)

            message = ""
            # log.debug('adding new data to the %s table; query: %s' %
            # (dbTableName, addValue))
            message = writequery(
                log=log,
                sqlQuery=insertCommand,
                dbConn=dbConn,
                Force=True,
                manyValueList=theseInserts
            )

        if message == "unknown column":
            for aDict in batch:
                convert_dictionary_to_mysql_table(
                    dbConn=dbConn,
                    log=log,
                    dictionary=aDict,
                    dbTableName=dbTableName,
                    uniqueKeyList=uniqueKeyList,
                    dateModified=dateModified,
                    reDatetime=reDate,
                    replace=replace
                )
        else:
            inserted = True

        dbConn.commit()

    log.debug('completed the ``_insert_single_batch_into_database`` function')
    return "None"


def _add_dictlist_to_database_via_load_in_file(
        masterListIndex,
        log,
        dbTablename,
        dbSettings,
        dateModified=False):
    """*load a list of dictionaries into a database table with load data infile*

    **Key Arguments:**

        - ``masterListIndex`` -- the index of the sharedList of dictionary lists to process
        - ``dbTablename`` -- the name of the database table to add the list to
        - ``dbSettings`` -- the dictionary of database settings
        - ``log`` -- logger
        - ``dateModified`` -- add a dateModified stamp with an updated flag to rows?

    **Return:**
        - None

    **Usage:**
        .. todo::

            add usage info
            create a sublime snippet for usage

        .. code-block:: python

            usage code
    """
    log.debug('starting the ``_add_dictlist_to_database_via_load_in_file`` function')

    global sharedList

    dictList = sharedList[masterListIndex][0]

    count = sharedList[masterListIndex][1]
    if count > totalCount:
        count = totalCount
    ltotalCount = totalCount

    # SETUP ALL DATABASE CONNECTIONS
    dbConn = database(
        log=log,
        dbSettings=dbSettings
    ).connect()

    now = datetime.now()
    tmpTable = now.strftime("tmp_%Y%m%dt%H%M%S%f")

    # CREATE A TEMPORY TABLE TO ADD DATA TO
    sqlQuery = """CREATE TEMPORARY TABLE %(tmpTable)s SELECT * FROM %(dbTablename)s WHERE 1=0;""" % locals()
    writequery(
        log=log,
        sqlQuery=sqlQuery,
        dbConn=dbConn
    )

    csvColumns = [k for d in dictList for k in d.keys()]
    csvColumns = list(set(csvColumns))
    csvColumnsString = (', ').join(csvColumns)

    df = pd.DataFrame(dictList)
    df.replace(['nan', 'None', '', 'NaN', np.nan], '\\N', inplace=True)
    df.to_csv('/tmp/%(tmpTable)s' % locals(), sep="|",
              index=False, escapechar="\\", quotechar='"', columns=csvColumns, encoding='utf-8')

    sqlQuery = """LOAD DATA LOCAL INFILE '/tmp/%(tmpTable)s'
INTO TABLE %(tmpTable)s
FIELDS TERMINATED BY '|' OPTIONALLY ENCLOSED BY '"'
IGNORE 1 LINES
(%(csvColumnsString)s);""" % locals()

    writequery(
        log=log,
        sqlQuery=sqlQuery,
        dbConn=dbConn
    )

    updateStatement = ""
    for i in csvColumns:
        updateStatement += "`%(i)s` = VALUES(`%(i)s`), " % locals()
    if dateModified:
        updateStatement += "dateLastModified = NOW(), updated = 1"
    else:
        updateStatement = updateStatement[0:-2]

    sqlQuery = """
INSERT IGNORE INTO %(dbTablename)s
SELECT * FROM %(tmpTable)s
ON DUPLICATE KEY UPDATE %(updateStatement)s;""" % locals()
    writequery(
        log=log,
        sqlQuery=sqlQuery,
        dbConn=dbConn
    )

    sqlQuery = """DROP TEMPORARY TABLE %(tmpTable)s;""" % locals()
    writequery(
        log=log,
        sqlQuery=sqlQuery,
        dbConn=dbConn
    )

    try:
        os.remove('/tmp/%(tmpTable)s' % locals())
    except:
        pass

    log.debug(
        'completed the ``_add_dictlist_to_database_via_load_in_file`` function')
    return None

# use the tab-trigger below for new function
# xt-def-function



#!/usr/local/bin/python
# encoding: utf-8
"""
*Probe a database to determine if a given table exists*

:Author:
    David Young

:Date Created:
    June 21, 2016
"""
################# GLOBAL IMPORTS ####################
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from fundamentals.mysql import readquery


def table_exists(
        dbConn,
        log,
        dbTableName):
    """*Probe a database to determine if a given table exists*

    **Key Arguments:**
        - ``dbConn`` -- mysql database connection
        - ``log`` -- logger
        - ``dbTableName`` -- the database tablename

    **Return:**
        - ``tableExists`` -- True or False

    **Usage:**

        To test if a table exists in a database:

        .. code-block:: python 

            from fundamentals.mysql import table_exists
            exists = table_exists(
                dbConn=dbConn,
                log=log,
                dbTableName="stupid_named_table"
            )

            print exists

            # OUTPUT: False
    """
    log.debug('starting the ``table_exists`` function')

    sqlQuery = u"""
        SELECT count(*)
        FROM information_schema.tables
        WHERE table_name = '%(dbTableName)s'
    """ % locals()
    tableExists = readquery(
        log=log,
        sqlQuery=sqlQuery,
        dbConn=dbConn,
        quiet=False
    )

    if tableExists[0]["count(*)"] == 0:
        tableExists = False
    else:
        tableExists = True

    log.debug('completed the ``table_exists`` function')
    return tableExists



#!/usr/local/bin/python
# encoding: utf-8
"""
*A function to quickly add multiprocessing to any program*

:Author:
    David Young

:Date Created:
    November  9, 2017
"""
################# GLOBAL IMPORTS ####################
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from multiprocess import cpu_count, Pool
from functools import partial
import inspect
import psutil


def fmultiprocess(
        log,
        function,
        inputArray,
        poolSize=False,
        timeout=3600,
        **kwargs):
    """multiprocess pool

    **Key Arguments:**
        - ``log`` -- logger
        - ``function`` -- the function to multiprocess
        - ``inputArray`` -- the array to be iterated over
        - ``poolSize`` -- limit the number of CPU that are used in multiprocess job
        - ``timeout`` -- time in sec after which to raise a timeout error if the processes have not completed

    **Return:**
        - ``resultArray`` -- the array of results

    **Usage:**

        .. code-block:: python 

            from fundamentals import multiprocess
            # DEFINE AN INPUT ARRAY
            inputArray = range(10000)
            results = multiprocess(log=log, function=functionName, poolSize=10, timeout=300,
                                  inputArray=inputArray, otherFunctionKeyword="cheese")
    """
    log.debug('starting the ``multiprocess`` function')

    # DEFINTE POOL SIZE - NUMBER OF CPU CORES TO USE (BEST = ALL - 1)
    if not poolSize:
        poolSize = psutil.cpu_count()

    if poolSize:
        p = Pool(processes=poolSize)
    else:
        p = Pool()

    cpuCount = psutil.cpu_count()
    chunksize = int((len(inputArray) + 1) / cpuCount * 3)

    if chunksize == 0:
        chunksize = 1

    # MAP-REDUCE THE WORK OVER MULTIPLE CPU CORES
    if "log" in inspect.getargspec(function)[0]:
        mapfunc = partial(function, log=log, **kwargs)
        resultArray = p.map_async(mapfunc, inputArray, chunksize=chunksize)
    else:
        mapfunc = partial(function, **kwargs)
        resultArray = p.map_async(mapfunc, inputArray, chunksize=chunksize)

    resultArray = resultArray.get(timeout=timeout)

    p.close()
    p.terminate()

    log.debug('completed the ``multiprocess`` function')
    return resultArray



import os
import nose
import shutil
import yaml
import unittest
from fundamentals.utKit import utKit
from fundamentals import tools

su = tools(
    arguments={"settingsFile": None},
    docString=__doc__,
    logLevel="DEBUG",
    options_first=False,
    projectName="fundamentals"
)
arguments, settings, log, dbConn = su.setup()

# # load settings
# stream = file(
#     "/Users/Dave/.config/rockAtlas/rockAtlas.yaml", 'r')
# settings = yaml.load(stream)
# stream.close()

# SETUP AND TEARDOWN FIXTURE FUNCTIONS FOR THE ENTIRE MODULE
moduleDirectory = os.path.dirname(__file__)
utKit = utKit(moduleDirectory)
log, dbConn, pathToInputDir, pathToOutputDir = utKit.setupModule()
utKit.tearDownModule()

# load settings
stream = file(
    pathToInputDir + "/example_settings_2.yaml", 'r')
settings = yaml.load(stream)
stream.close()

import shutil
try:
    shutil.rmtree(pathToOutputDir)
except:
    pass
# COPY INPUT TO OUTPUT DIR
shutil.copytree(pathToInputDir, pathToOutputDir)

# Recursively create missing directories
if not os.path.exists(pathToOutputDir):
    os.makedirs(pathToOutputDir)

# xt-setup-unit-testing-files-and-folders

dbSettings = settings["database settings"]["atlasMovers"]
dbSettings2 = settings["database settings"]["dryx_unit_testing"]


class test_database(unittest.TestCase):

    def test_tunnel(self):

        from fundamentals.mysql import database
        db = database(
            log=log,
            dbSettings=dbSettings
        )
        sshPort = db._setup_tunnel(
            tunnelParameters=dbSettings["tunnel"]
        )

        return

    def test_no_tunnel(self):

        from fundamentals.mysql import database
        db = database(
            log=log,
            dbSettings=dbSettings2
        ).connect()

        return

    def test_database_function(self):

        # SETUP ALL DATABASE CONNECTIONS
        from fundamentals.mysql import database
        dbConn = database(
            log=log,
            dbSettings=dbSettings
        ).connect()

        from fundamentals.mysql import readquery
        sqlQuery = u"""
            SELECT VERSION();
        """ % locals()
        rows = readquery(
            log=log,
            sqlQuery=sqlQuery,
            dbConn=dbConn,
            quiet=False
        )
        print rows

    def test_database_function_exception(self):

        from fundamentals.mysql import database
        try:
            this = database(
                log=log,
                dbSettings=dbSettings,
                fakeKey="break the code"
            )
            this.get()
            assert False
        except Exception, e:
            assert True
            print str(e)

        # x-print-testpage-for-pessto-marshall-web-object

    # x-class-to-test-named-worker-function



#!/usr/local/bin/python
# encoding: utf-8
"""
*Given a database connection and a database table name, return the column names for the table*

:Author:
    David Young

:Date Created:
    June 21, 2016
"""
################# GLOBAL IMPORTS ####################
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from fundamentals.mysql import readquery


def get_database_table_column_names(
    dbConn,
    log,
    dbTable
):
    """get database table column names

    **Key Arguments:**
        - ``dbConn`` -- mysql database connection
        - ``log`` -- logger
        - ``dbTable`` -- database tablename

    **Return:**
        - ``columnNames`` -- table column names

    **Usage:**

        To get the column names of a table in a given database:

        .. code-block:: python 

            from fundamentals.mysql import get_database_table_column_names
            columnNames = get_database_table_column_names(
                dbConn=dbConn,
                log=log,
                dbTable="test_table"
            )
    """
    log.debug('starting the ``get_database_table_column_names`` function')

    sqlQuery = """SELECT * FROM %s LIMIT 1""" \
        % (dbTable, )
    # ############### >ACTION(S) ################
    try:
        rows = readquery(
            log=log,
            sqlQuery=sqlQuery,
            dbConn=dbConn,
        )
    except Exception as e:
        log.error(
            'could not find column names for dbTable %s - failed with this error: %s ' %
            (dbTable, str(e)))
        return -1
    columnNames = rows[0].keys()

    log.debug('completed the ``get_database_table_column_names`` function')
    return columnNames



#!/usr/local/bin/python
# encoding: utf-8
"""
*Take key-values from a yaml file including a tablename(s) and add them to a mysql database table*

Usage:
    yaml2db [-d] -s <pathToSettingsFile> <pathToYaml>
    yaml2db [-d] --host=<host> --user=<user> --passwd=<passwd> --dbName=<dbName> <pathToYaml>

Options:

    pathToYaml            path to a single yaml file or directory of yaml files
    pathToSettingsFile    path to a settings file with logging and database information (yaml file)
    --host=<host>         the database host
    --user=<user>         database user
    --passwd=<passwd>     database user password
    --dbName=<dbName>     name of the database to add the table content to

    -d, --delete          delete yaml file(s) once added to datbase
    -h, --help            show this help message
    -v, --version         show version
    -s, --settings        the settings file
"""

import sys
import os
import yaml
import requests
import re
import glob
import docopt
from fundamentals import tools, times
from fundamentals.mysql import convert_dictionary_to_mysql_table
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()


def main(arguments=None):
    """
    The main function used when ``yaml_to_database.py`` when installed as a cl tool
    """

    # setup the command-line util settings
    su = tools(
        arguments=arguments,
        docString=__doc__,
        logLevel="WARNING",
        options_first=False,
        projectName=False
    )
    arguments, settings, log, dbConn = su.setup()

    # unpack remaining cl arguments using `exec` to setup the variable names
    # automatically
    for arg, val in arguments.iteritems():
        if arg[0] == "-":
            varname = arg.replace("-", "") + "Flag"
        else:
            varname = arg.replace("<", "").replace(">", "")
        if isinstance(val, str) or isinstance(val, unicode):
            exec(varname + " = '%s'" % (val,))
        else:
            exec(varname + " = %s" % (val,))
        if arg == "--dbConn":
            dbConn = val
        log.debug('%s = %s' % (varname, val,))

    if os.path.isfile(pathToYaml):
        from fundamentals.mysql import yaml_to_database
        # PARSE YAML FILE CONTENTS AND ADD TO DATABASE
        yaml2db = yaml_to_database(
            log=log,
            settings=settings,
            dbConn=dbConn
        )
        yaml2db.add_yaml_file_content_to_database(
            filepath=pathToYaml,
            deleteFile=deleteFlag
        )
        basename = os.path.basename(pathToYaml)
        print "Content of %(basename)s added to database" % locals()

    else:
        from fundamentals.mysql import yaml_to_database
        yaml2db = yaml_to_database(
            log=log,
            settings=settings,
            dbConn=dbConn,
            pathToInputDir=pathToYaml,
            deleteFiles=deleteFlag
        )
        yaml2db.ingest()
        print "Content of %(pathToYaml)s directory added to database" % locals()

    return


class yaml_to_database():
    """
    *Take key-values from yaml files including a tablename(s) and add them to a mysql database table*

    **Key Arguments:**
        - ``log`` -- logger
        - ``settings`` -- the settings dictionary
        - ``pathToInputDir`` -- path to the directory containing the yaml files that will be added to the database table(s). Default *False*
        - ``dbConn`` -- connection to database to add the content to
        - ``deleteFiles`` - - delete the yamls files once their content has been added to the database. Default * False*

    **Usage:**

        To setup your logger, settings and database connections, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_). 

        To initiate a ``yaml2db`` object, use the following:

        .. code-block:: python 

            from fundamentals.mysql import yaml_to_database
            yaml2db = yaml_to_database(
                log=log,
                settings=settings,
                dbConn=dbConn,
                pathToInputDir="/path/to/yaml/directory",
                deleteFiles=False
            ) 

        And here's an example of the content in a yaml file that this ``yaml2db`` object can parse:

        .. code-block:: yaml

            title: Why you should do most of your text editing in : Sublime Text | Sublime Text Tips
            url: http://sublimetexttips.com/why-you-should-do-most-of-your-text-editing-in-sublime-text/?utm_source=drip&utm_medium=email&utm_campaign=editor-proliferation
            kind: webpage
            subtype: article
            table: web_articles,podcasts 
    """
    # Initialisation

    def __init__(
            self,
            log,
            dbConn,
            pathToInputDir=False,
            settings=False,
            deleteFiles=False

    ):
        self.log = log
        log.debug("instansiating a new 'yaml_to_database' object")
        self.settings = settings
        self.pathToInputDir = pathToInputDir
        self.dbConn = dbConn
        self.deleteFiles = deleteFiles
        # xt-self-arg-tmpx

        return None

    def ingest(self):
        """
        *ingest the contents of the directory of yaml files into a database*

        **Return:**
            - None

        **Usage:**

            To import an entire directory of yaml files into a database, use the following:

        .. code-block:: python 

            from fundamentals.mysql import yaml_to_database
            yaml2db = yaml_to_database(
                log=log,
                settings=settings,
                dbConn=dbConn,
                pathToInputDir="/path/to/yaml/directory",
                deleteFiles=False
            ) 
            yaml2db.ingest() 
        """
        self.log.debug('starting the ``ingest`` method')

        for d in os.listdir(self.pathToInputDir):
            if os.path.isfile(os.path.join(self.pathToInputDir, d)) and "yaml" in d.lower():
                self.add_yaml_file_content_to_database(
                    filepath=os.path.join(self.pathToInputDir, d),
                    deleteFile=self.deleteFiles
                )

        self.log.debug('completed the ``ingest`` method')
        return None

    def add_yaml_file_content_to_database(
        self,
        filepath,
        deleteFile=False
    ):
        """*given a file to a yaml file, add yaml file content to database*

        **Key Arguments:**
            - ``filepath`` -- the path to the yaml file
            - ``deleteFile`` -- delete the yaml file when its content has been added to the database. Default *False*

        **Return:**
            - None

        **Usage:**

            To parse and import the contents of a single yaml file into the database, use the following:

            .. code-block:: python 

                from fundamentals.mysql import yaml_to_database
                # PARSE YAML FILE CONTENTS AND ADD TO DATABASE
                yaml2db = yaml_to_database(
                    log=log,
                    settings=settings,
                    dbConn=dbConn
                ) 
                yaml2db.add_yaml_file_content_to_database(
                    filepath=${1:"/path/to/file.yaml"},
                    deleteFile=True
                )
        """
        self.log.debug(
            'completed the ````add_yaml_file_content_to_database`` method')

        import codecs
        try:
            self.log.debug("attempting to open the file %s" % (filepath,))
            readFile = codecs.open(filepath, encoding='utf-8', mode='r')
            thisData = readFile.read()
            readFile.close()
        except IOError, e:
            message = 'could not open the file %s' % (filepath,)
            self.log.critical(message)
            raise IOError(message)
        readFile.close()

        matchObject = re.finditer(
            r'(^|\n)(?P<key>[^\:]*)\:\s(?P<value>.*?)(\n|$)',
            thisData,
            flags=re.M | re.S  # re.S
        )

        yamlContent = {}
        for match in matchObject:
            if match.group("value")[0] == '"' and match.group("value")[-1] == '"':
                v = match.group("value")[1:-1]
            elif match.group("value")[0] == "'" and match.group("value")[-1] == "'":
                v = match.group("value")[1:-1]
            else:
                v = match.group("value")
            yamlContent[match.group("key")] = v

        if "table" not in yamlContent:
            self.log.warning(
                'A table value is need in the yaml content to indicate which database table to add the content to: %(filepath)s' % locals())
            return None

        # NOTE THERE MAY BE MORE THAN ONE DATABASE TABLE
        dbTablesTmp = yamlContent["table"].split(",")
        del yamlContent["table"]
        dbTables = []
        dbTables[:] = [d.strip() for d in dbTablesTmp]

        # UNSHORTEN URL
        try:
            r = requests.head(yamlContent["url"], allow_redirects=True)
            yamlContent["url"] = r.url
        except:
            pass

        yamlContent["original_yaml_path"] = filepath

        if "url" in yamlContent:
            uniqueKeyList = ["url"]
        else:
            uniqueKeyList = []

        for t in dbTables:
            convert_dictionary_to_mysql_table(
                dbConn=self.dbConn,
                log=self.log,
                dictionary=yamlContent,
                dbTableName=t,
                uniqueKeyList=uniqueKeyList,
                dateModified=True,
                returnInsertOnly=False,
                replace=True
            )
        if deleteFile:
            os.remove(filepath)

        self.log.debug(
            'completed the ``add_yaml_file_content_to_database`` method')
        return None

    # use the tab-trigger below for new method
    # xt-class-method


if __name__ == '__main__':
    main()



#!/usr/local/bin/python
# encoding: utf-8
"""
*Given a directory of MySQL scripts, execute the scripts and process the script files according to their success or failure*

Run the following code once to set a login-path for your mysql server:

    mysql_config_editor set --login-path=<uniqueLoginName> --host=localhost --user=<myUsername> --password

This store's your credentials in an encrypted file located at '~/.mylogin.cnf'. 
Use `mysql_config_editor print --all` to see all of the login-paths set.


Usage:
    mysqlSucker <pathToDirectory> <loginPath> <databaseName> [-s successRule -f failureRule]

    pathToDirectory       path to the directory containing the sql scripts to run (scripts must have `.sql` extension)
    loginPath             the local-path as set with `mysql_config_editor` (`mysqlSucker -h` for more details)
    databaseName          the name of the database to execute the scripts within  
            
Options:
    -h, --help                                  show this help message
    -s successRule, --success successRule       what to do if script succeeds. Default *None* [None|delete|subFolderName]
    -f failureRule, --failure failureRule       what to do if script fails. Default *None* [None|delete|subFolderName]  

:Examples:

    To simply execute the scripts in a directory you can run:

        mysqlSucker /path/to/scriptdir myLoginPath myDatabaseName

    To delete script after thay have executed successfully:

        mysqlSucker /path/to/scriptdir myLoginPath myDatabaseName -s delete

    To move successful script to a `passed` sub-directory of `/path/to/scriptdir` and failed scripts to a `failed` sub-directory

        mysqlSucker /path/to/scriptdir myLoginPath myDatabaseName -s pass -f failed

:Author:
    David Young

:Date Created:
    September 22, 2016
"""
################# GLOBAL IMPORTS ####################
import sys
import os
import time
import collections
import datetime
from subprocess import Popen, PIPE, STDOUT
os.environ['TERM'] = 'vt100'
from fundamentals import tools


def directory_script_runner(
        log,
        pathToScriptDirectory,
        databaseName,
        loginPath,
        force=True,
        waitForResult=True,
        successRule=None,
        failureRule=None):
    """A function to run all of the mysql scripts in a given directory (in a modified date order, oldest first) and then act on the script file in accordance with the succcess or failure of its execution

    As it's insecure to pass in mysql database credentials via the command-line, run the following command from the terminal 

    .. code-block:: bash

        mysql_config_editor set --login-path=<uniqueLoginName> --host=localhost --user=<myUsername> --password
        > Enter password:

    This will store your credentials in an encrypted file located at '~/.mylogin.cnf'. This function takes advantage of the `--login-path` so as not to compromise the user's credentials. Use `mysql_config_editor print --all` to see all of the login-paths set.

    **Key Arguments:**
        - ``log`` -- logger
        - ``pathToScriptDirectory`` -- the path to the directory containing the sql script to be run
        - ``databaseName`` -- the name of the database 
        - ``force`` -- force the script to run, skipping over lines with errors, Default *True*
        - ``loginPath`` -- the local-path as set with `mysql_config_editor`
        - ``waitForResult`` -- wait for the mysql script to finish execution? If 'False' the MySQL script will run in background (do not wait for completion), or if 'delete' the script will run then delete regardless of success status. Default *True*. [True|False|delete]
        - ``successRule`` -- what to do if script succeeds. Default *None* [None|delete|subFolderName]
        - ``failureRule`` -- what to do if script fails. Default *None* [None|delete|subFolderName]

    **Return:**
        - None

    **Usage:**

        To run the scripts in the directroy and not act on the script file use something similar to:

        .. code-block:: python 

            from fundamentals.mysql import directory_script_runner
            directory_script_runner(
                log=log,
                pathToScriptDirectory="/path/to/mysql_scripts",
                databaseName="imports",
                loginPath="myLoginDetails"
            )

        To delete successful scripts and archive failed scripts for later inspection:

        .. code-block:: python 

            from fundamentals.mysql import directory_script_runner
            directory_script_runner(
                log=log,
                pathToScriptDirectory="/path/to/mysql_scripts",
                databaseName="imports",
                loginPath="myLoginDetails",
                successRule="delete",
                failureRule="failed"
            )

        This creates a folder at `/path/to/mysql_scripts/failed` and moves the failed scripts into that folder.

        Finally to execute the scripts within a directory but not wait for the results to return (much fast but you lose error checking in the MySQL scripts):

        .. code-block:: python 

            from fundamentals.mysql import directory_script_runner
            directory_script_runner(
                log=log,
                pathToScriptDirectory="/path/to/mysql_scripts",
                databaseName="imports",
                loginPath="myLoginDetails",
                waitForResults=False
            )

        Setting ``waitForResults`` = 'delete' will trash the script once it has run (or failed ... be very careful!)
    """
    log.debug('starting the ``directory_script_runner`` function')

    # COMPILE A DICTIONARY OF SCRIPTS / MODIFIED TIMES
    scriptList = {}
    for d in os.listdir(pathToScriptDirectory):
        filePath = os.path.join(pathToScriptDirectory, d)
        filename = os.path.basename(filePath)
        extension = filePath.split(".")[-1]
        if os.path.isfile(filePath) and extension == "sql":
            modified = datetime.datetime.strptime(time.ctime(
                os.path.getmtime(filePath)), "%a %b %d %H:%M:%S %Y")
            scriptList[str(modified) + filename] = filePath

    # FORCE MYSQL SCRIPT?
    if force:
        force = "--force"
    else:
        force = ""

    # ORDER THE DICTIONARY BY MODIFIED TIME - OLDEST FIRST
    scriptList = collections.OrderedDict(sorted(scriptList.items()))
    for k, v in scriptList.iteritems():
        scriptname = os.path.basename(v)
        if waitForResult == True:
            cmd =  """mysql --login-path=%(loginPath)s %(force)s %(databaseName)s < "%(v)s" """ % locals(
            )
            p = Popen(cmd, stdout=PIPE, stderr=PIPE, close_fds=True,
                      env={'PATH': os.getenv('PATH') + ":/usr/local/bin:/usr/bin:", "MYSQL_TEST_LOGIN_FILE": os.getenv('HOME') + "/.mylogin.cnf"}, shell=True)
            stdout, stderr = p.communicate()

            if len(stderr):
                log.error(
                    "MySQL Script `%(scriptname)s` Failed: '%(stderr)s'" % locals())
                if failureRule == None or failureRule == False:
                    pass
                elif failureRule == "delete":
                    os.remove(v)
                elif "/" not in failureRule:
                    moveTo = pathToScriptDirectory + "/" + failureRule
                    # Recursively create missing directories
                    if not os.path.exists(moveTo):
                        os.makedirs(moveTo)
                    moveTo = moveTo + "/" + scriptname
                    try:
                        log.debug("attempting to rename file %s to %s" %
                                  (v, moveTo))
                        os.rename(v, moveTo)
                    except Exception, e:
                        log.error(
                            "could not rename file %s to %s - failed with this error: %s " % (v, moveTo, str(e),))
            else:
                if successRule == None or successRule == False:
                    pass
                elif successRule == "delete":
                    os.remove(v)
                elif "/" not in successRule:
                    moveTo = pathToScriptDirectory + "/" + successRule
                    # Recursively create missing directories
                    if not os.path.exists(moveTo):
                        os.makedirs(moveTo)
                    moveTo = moveTo + "/" + scriptname
                    try:
                        log.debug("attempting to rename file %s to %s" %
                                  (v, moveTo))
                        os.rename(v, moveTo)
                    except Exception, e:
                        log.error(
                            "could not rename file %s to %s - failed with this error: %s " % (v, moveTo, str(e),))
        else:
            if waitForResult == "delete":
                cmd =  """mysql --login-path=%(loginPath)s %(force)s  %(databaseName)s < "%(v)s" > /dev/null 2>&1 & rm "%(v)s" """ % locals()
            else:
                cmd =  """mysql --login-path=%(loginPath)s %(force)s  %(databaseName)s < "%(v)s" > /dev/null 2>&1 """ % locals()
            p = Popen(cmd, close_fds=True,
                      env={'PATH': os.getenv('PATH') + ":/usr/local/bin:/usr/bin:", "MYSQL_TEST_LOGIN_FILE": os.getenv('HOME') + "/.mylogin.cnf"}, shell=True, stdin=None, stdout=None, stderr=None)

    log.debug('completed the ``directory_script_runner`` function')
    return None


def main(arguments=None):
    """
    The main function used when ``directory_script_runner.py`` is run as a single script from the cl, or when installed as a cl command
    """

    # setup the command-line util settings
    su = tools(
        arguments=arguments,
        docString=__doc__,
        logLevel="WARNING",
        options_first=False,
        projectName="fundmentals"
    )
    arguments, settings, log, dbConn = su.setup()

    # UNPACK REMAINING CL ARGUMENTS USING `EXEC` TO SETUP THE VARIABLE NAMES
    # AUTOMATICALLY
    for arg, val in arguments.iteritems():
        if arg[0] == "-":
            varname = arg.replace("-", "") + "Flag"
        else:
            varname = arg.replace("<", "").replace(">", "")
        if isinstance(val, str) or isinstance(val, unicode):
            exec(varname + " = '%s'" % (val,))
        else:
            exec(varname + " = %s" % (val,))
        if arg == "--dbConn":
            dbConn = val
        log.debug('%s = %s' % (varname, val,))

    if successFlag and successFlag.lower() == "none":
        successFlag = None
    if failureFlag and failureFlag.lower() == "none":
        failureFlag = None

    directory_script_runner(
        log=log,
        pathToScriptDirectory=pathToDirectory,
        databaseName=databaseName,
        loginPath=loginPath,
        successRule=successFlag,
        failureRule=failureFlag
    )

    return

if __name__ == '__main__':
    main()
